@@ -1,15 +1,27 @@
 {-# LANGUAGE AllowAmbiguousTypes #-}
 {-# LANGUAGE DerivingVia #-}
 {-# LANGUAGE MagicHash #-}
+{-# LANGUAGE PatternSynonyms #-}
+{-# LANGUAGE QuantifiedConstraints #-}
 {-# LANGUAGE RoleAnnotations #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE UnboxedTuples #-}
+{-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE UnliftedNewtypes #-}
 {-# OPTIONS_HADDOCK not-home #-}
 
 module Bluefin.Internal where
 
 import qualified Bluefin.Internal.Exception.Scoped as ScopedException
+import Bluefin.Internal.OneWayCoercible
+  ( OneWayCoercible (oneWayCoercibleImpl),
+    OneWayCoercion,
+    gOneWayCoercible,
+    oneWayCoerce,
+    oneWayCoerceWith,
+    oneWayCoercible,
+    unsafeOneWayCoercible,
+  )
 import qualified Control.Concurrent.Async as Async
 import Control.Concurrent.MVar (newEmptyMVar, putMVar, takeMVar)
 import qualified Control.Exception
@@ -26,6 +38,7 @@ import Data.Kind (Type)
 import Data.Proxy (Proxy (Proxy))
 import Data.Type.Coercion (Coercion (Coercion))
 import GHC.Exts (Proxy#, proxy#)
+import GHC.Generics (Generic)
 import System.IO.Unsafe (unsafePerformIO)
 import Unsafe.Coerce (unsafeCoerce)
 import Prelude hiding (drop, head, read, return)
@@ -39,12 +52,18 @@ infixr 9 :&
 
 type (:&) = Union
 
-type role Eff nominal representational
-
 newtype Eff (es :: Effects) a = UnsafeMkEff {unsafeUnEff :: IO a}
   deriving stock (Functor)
   deriving newtype (Applicative, Monad)
 
+type role Eff nominal representational
+
+instance (e :> es) => OneWayCoercible (Eff e) (Eff es) where
+  oneWayCoercibleImpl = oneWayCoercible
+
+instance (e :> es) => OneWayCoercible (Eff e r) (Eff es r) where
+  oneWayCoercibleImpl = oneWayCoercible
+
 -- | Because doing 'IO' operations inside 'Eff' requires a value-level
 -- argument we can't give @IO@-related instances to @Eff@ directly.
 -- Instead we wrap it in @EffReader@.
@@ -264,7 +283,7 @@ weakenEff :: t `In` t' -> Eff t r -> Eff t' r
 weakenEff _ = unsafeCoerceEff
 
 insertFirst :: Eff b r -> Eff (c1 :& b) r
-insertFirst = weakenEff (drop (eq (# #)))
+insertFirst = weakenEff (drop (eq ZW))
 
 insertSecond :: Eff (c1 :& b) r -> Eff (c1 :& (c2 :& b)) r
 insertSecond = insertManySecond
@@ -273,13 +292,13 @@ insertManySecond :: (b :> c) => Eff (c1 :& b) r -> Eff (c1 :& c) r
 insertManySecond = weakenEff (bimap has has)
 
 assoc1Eff :: Eff ((a :& b) :& c) r -> Eff (a :& (b :& c)) r
-assoc1Eff = weakenEff (assoc1 (# #))
+assoc1Eff = weakenEff (assoc1 ZW)
 
 pushFirst :: Eff a r -> Eff (a :& b) r
-pushFirst = weakenEff (fstI (# #))
+pushFirst = weakenEff (fstI ZW)
 
 mergeEff :: Eff (a :& a) r -> Eff a r
-mergeEff = weakenEff (merge (# #))
+mergeEff = weakenEff (merge ZW)
 
 inContext :: (e2 :> e1) => Eff (e1 :& e2) r -> Eff e1 r
 inContext = weakenEff (subsume1 has)
@@ -326,15 +345,29 @@ type role StateSource nominal
 -- | Handle to an exception of type @exn@
 newtype Exception exn (e :: Effects)
   = MkException (forall a. exn -> Eff e a)
+  deriving (Handle) via OneWayCoercibleHandle (Exception exn)
+
+type role Exception representational nominal
+
+instance (e :> es) => OneWayCoercible (Exception ex e) (Exception ex es) where
+  oneWayCoercibleImpl = oneWayCoercible
 
 -- | A handle to a strict mutable state of type @s@
 newtype State s (e :: Effects) = UnsafeMkState (IORef s)
+  deriving (Handle) via OneWayCoercibleHandle (State s)
 
 type role State representational nominal
 
+instance (e :> es) => OneWayCoercible (State s e) (State s es) where
+  oneWayCoercibleImpl = oneWayCoercible
+
 -- | A handle to a coroutine that yields values of type @a@ and then
 -- expects values of type @b@.
 newtype Coroutine a b (e :: Effects) = MkCoroutine (a -> Eff e b)
+  deriving (Handle) via OneWayCoercibleHandle (Coroutine a b)
+
+instance (e :> es) => OneWayCoercible (Coroutine a b e) (Coroutine a b es) where
+  oneWayCoercibleImpl = oneWayCoercible
 
 -- | A handle to a stream that yields values of type @a@.  It is
 -- implemented as a handle to a coroutine that yields values of type
@@ -355,72 +388,142 @@ type Consume a = Coroutine () a
 --
 -- @
 -- data Application e = MkApplication
---   { queryDatabase :: forall e'. String -> Int -> Eff (e' :& e) [String],
+--   { queryDatabase :: String -> Int -> Eff e [String],
 --     applicationState :: State (Int, Bool) e,
 --     logger :: Stream String e
 --   }
--- @
+--   deriving (Generic)
+--   deriving (Handle) via 'OneWayCoercibleHandle' Application
 --
--- To define @mapHandle@ for @Application@ you should apply
--- @mapHandle@ to all the fields that are themeselves handles and
--- apply 'useImplUnder' to all the fields that are dynamic effects:
---
--- @
--- instance Handle Application where
---   mapHandle
---     MkApplication
---       { queryDatabase = q,
---         applicationState = a,
---         logger = l
---       } =
---       MkApplication
---         { queryDatabase = \\s i -> useImplUnder (q s i),
---           applicationState = mapHandle a,
---           logger = mapHandle l
---         }
+-- instance (e :> es) => 'OneWayCoercible' (Application e) (Application es) where
+--   oneWayCoercibleImpl = 'gOneWayCoercible'
 -- @
 class Handle (h :: Effects -> Type) where
+  {-# MINIMAL handleImpl | mapHandle #-}
+  -- | Define @handleImpl@ using 'handleMapHandle'.
+  handleImpl :: HandleD h
+  handleImpl = MkHandleD mapHandle
+
   -- | Used to create compound effects, i.e. handles that contain
   -- other handles.
+  --
+  -- You should not define this method in your type classes. Instead
+  -- define @handleImpl@.  In a future version of Bluefin, @mapHandle@
+  -- will no longer be a method.  If you previously had a definition
+  -- of @mapHandle@, for example like this:
+  --
+  -- @
+  -- instance Handle MyHandle where
+  --   mapHandle h = \<definition\>
+  -- @
+  --
+  -- you should change it to
+  --
+  -- @
+  -- data MyHandle e = ...
+  --   deriving (Generic)
+  --   deriving (Handle) via OneWayCoercibleHandle MyHandle
+  --
+  -- instance (e :> es) => OneWayCoercible (MyHandle e) (MyHandle es) where
+  --   oneWayCoercibleImpl = gOneWayCoercible
+  -- @
+  --
+  -- If that doesn't work for any reason you can always reuse your old
+  -- definition of @mapHandle@ as follows.  However,
+  -- 'handleMapHandle' will be removed at some point in the future, so
+  -- you can [open a new issue on
+  -- Bluefin](https://github.com/tomjaguarpaw/bluefin/issues/new) to
+  -- ask for advice.
+  --
+  -- @
+  -- instance Handle MyHandle where
+  --   handleImpl = handleMapHandle $ \\h -> \<definition\>
+  -- @
   mapHandle :: (e :> es) => h e -> h es
+  mapHandle = case handleImpl of MkHandleD f -> f
 
-instance Handle (State s) where
-  mapHandle (UnsafeMkState s) = UnsafeMkState s
+-- | The type of the 'handleImpl' method of the 'Handle' class.
+-- Create a @HandleD@ using 'handleMapHandle'.
+newtype HandleD h = MkHandleD (forall e es. (e :> es) => h e -> h es)
 
-instance Handle (Exception s) where
-  mapHandle (MkException s) = MkException (weakenEff has . s)
+-- | For defining the 'handleImpl' method of the 'Handle' class.
+handleMapHandle ::
+  (forall e es. (e :> es) => h e -> h es) ->
+  -- | ͘
+  HandleD h
+handleMapHandle = MkHandleD
 
-instance Handle (Coroutine a b) where
-  mapHandle (MkCoroutine f) = MkCoroutine (fmap useImpl f)
+handleOneWayCoercible ::
+  (forall e es. (e :> es) => OneWayCoercible (h e) (h es)) =>
+  -- | ͘
+  HandleD h
+handleOneWayCoercible = MkHandleD oneWayCoerce
 
-instance Handle (Writer w) where
-  mapHandle (Writer wr) = Writer (mapHandle wr)
+handleOneWayCoercion ::
+  (forall e es. (e :> es) => OneWayCoercion (h e) (h es)) ->
+  -- | ͘
+  HandleD h
+handleOneWayCoercion owc = MkHandleD (oneWayCoerceWith owc)
 
-instance Handle IOE where
-  mapHandle MkIOE = MkIOE
+-- { OneWayCoercibleHandle
+
+-- | 'OneWayCoercibleHandle' is used to derive 'Handle' instances
+-- using @DerivingVia@
+newtype OneWayCoercibleHandle a es = MkOneWayCoercibleHandle (a es)
+  deriving stock (Generic)
+
+instance
+  forall h.
+  (forall e' es'. (e' :> es') => OneWayCoercible (OneWayCoercibleHandle h e') (OneWayCoercibleHandle h es')) =>
+  Handle (OneWayCoercibleHandle h)
+  where
+  handleImpl = handleOneWayCoercible
+
+instance
+  (OneWayCoercible (h e) (h es)) =>
+  OneWayCoercible (OneWayCoercibleHandle h e) (OneWayCoercibleHandle h es)
+  where
+  oneWayCoercibleImpl = gOneWayCoercible
+
+-- }
+
+-- | A convenience type whose only purpose is to avoid writing @(# #)@
+-- as an argument to functions which are only function because
+-- top-level definitions of unlifted kind are forbidden.
+newtype ZW = MkZW (# #)
+
+pattern ZW :: ZW
+pattern ZW <- MkZW (# #)
+  where
+    ZW = MkZW (# #)
+
+{-# COMPLETE ZW #-}
 
 newtype In (a :: Effects) (b :: Effects) = In# (# #)
 
-merge :: (# #) -> (a :& a) `In` a
-merge (# #) = In# (# #)
+unsafeInAxiom :: ZW -> e1 `In` e2
+unsafeInAxiom ZW = In# (# #)
 
-eq :: (# #) -> a `In` a
-eq (# #) = In# (# #)
+merge :: ZW -> (a :& a) `In` a
+merge ZW = In# (# #)
 
-fstI :: (# #) -> a `In` (a :& b)
-fstI (# #) = In# (# #)
+eq :: ZW -> a `In` a
+eq ZW = unsafeInAxiom ZW
 
-sndI :: (# #) -> a `In` (b :& a)
-sndI (# #) = In# (# #)
+fstI :: ZW -> a `In` (a :& b)
+fstI ZW = unsafeInAxiom ZW
+
+sndI :: ZW -> a `In` (b :& a)
+sndI ZW = unsafeInAxiom ZW
 
 cmp :: a `In` b -> b `In` c -> a `In` c
-cmp (In# (# #)) (In# (# #)) = In# (# #)
+cmp (In# (# #)) (In# (# #)) = unsafeInAxiom ZW
 
 bimap :: a `In` b -> c `In` d -> (a :& c) `In` (b :& d)
-bimap (In# (# #)) (In# (# #)) = In# (# #)
+bimap (In# (# #)) (In# (# #)) = unsafeInAxiom ZW
 
-assoc1 :: (# #) -> ((a :& b) :& c) `In` (a :& (b :& c))
-assoc1 (# #) = In# (# #)
+assoc1 :: ZW -> ((a :& b) :& c) `In` (a :& (b :& c))
+assoc1 ZW = unsafeInAxiom ZW
 
 drop :: a `In` b -> a `In` (c :& b)
 drop h = w2 (b h)
@@ -429,22 +532,22 @@ here :: a `In` b -> (a `In` (b :& c))
 here h = w (b2 h)
 
 w :: (a :& b) `In` c -> (a `In` c)
-w = cmp (fstI (# #))
+w = cmp (fstI ZW)
 
 w2 :: (b :& a) `In` c -> (a `In` c)
-w2 = cmp (sndI (# #))
+w2 = cmp (sndI ZW)
 
 b2 :: (a `In` b) -> ((a :& c) `In` (b :& c))
-b2 h = bimap h (eq (# #))
+b2 h = bimap h (eq ZW)
 
 b :: (a `In` b) -> (c :& a) `In` (c :& b)
-b = bimap (eq (# #))
+b = bimap (eq ZW)
 
 subsume1 :: (e2 `In` e1) -> (e1 :& e2) `In` e1
-subsume1 i = cmp (bimap (eq (# #)) i) (merge (# #))
+subsume1 i = cmp (bimap (eq ZW) i) (merge ZW)
 
 subsume2 :: (e1 `In` e2) -> (e1 :& e2) `In` e2
-subsume2 i = cmp (bimap i (eq (# #))) (merge (# #))
+subsume2 i = cmp (bimap i (eq ZW)) (merge ZW)
 
 -- | Effect subset constraint
 class (es1 :: Effects) :> (es2 :: Effects)
@@ -452,10 +555,16 @@ class (es1 :: Effects) :> (es2 :: Effects)
 -- | A set of effects @e@ is a subset of itself
 instance e :> e
 
+instanceProof1 :: ZW -> e `In` e
+instanceProof1 ZW = eq ZW
+
 -- | If @e@ is subset of @es@ then @e@ is a subset of a larger set, @x
 -- :& es@
 instance {-# INCOHERENT #-} (e :> es) => e :> (x :& es)
 
+instanceProof2 :: e `In` es -> e `In` (x :& es)
+instanceProof2 = drop
+
 -- Do we want this?
 -- instance {-# incoherent #-} (e :> es) => (e' :& e) :> (e' :> es)
 
@@ -464,12 +573,18 @@ instance {-# INCOHERENT #-} (e :> es) => e :> (x :& es)
 -- | @e@ is a subset of a larger set @e :& es@
 instance e :> (e :& es)
 
+instanceProof3 :: ZW -> e `In` (e :& es)
+instanceProof3 ZW = fstI ZW
+
 subset ::
   forall e1 es m.
   (Monad m) =>
   (e1 :> es) =>
   m ()
-subset = pure ()
+subset = satisfied @(e1 :> es)
+
+satisfied :: forall c m. (Monad m, c) => m ()
+satisfied = pure ()
 
 effTag :: Eff es (Proxy es)
 effTag = pure Proxy
@@ -499,7 +614,9 @@ throw ::
 throw h = case mapHandle h of MkException throw_ -> throw_
 
 has :: forall a b. (a :> b) => a `In` b
-has = In# (# #)
+-- This is safe because, as shown by instanceProof1/2/3, the only way
+-- to construct `a :> b` is if `a `In` b`.
+has = unsafeInAxiom ZW
 
 data Dict c where
   Dict :: forall c. (c) => Dict c
@@ -1217,9 +1334,13 @@ unwrap j = \case
 
 -- | Handle that allows you to run 'IO' operations
 data IOE (e :: Effects) = MkIOE
+  deriving (Handle) via OneWayCoercibleHandle IOE
 
 type role IOE nominal
 
+instance (e :> es) => OneWayCoercible (IOE e) (IOE es) where
+  oneWayCoercibleImpl = unsafeOneWayCoercible
+
 -- | Run an 'IO' operation in 'Eff'
 --
 -- @
@@ -1304,6 +1425,10 @@ head' c = do
     Left (l, _) -> Left l
 
 newtype Writer w e = Writer (Stream w e)
+  deriving (Handle) via OneWayCoercibleHandle (Writer w)
+
+instance (e :> es) => OneWayCoercible (Writer w e) (Writer w es) where
+  oneWayCoercibleImpl = oneWayCoercible
 
 -- |
 -- @
@@ -1360,6 +1485,9 @@ tell (Writer y) = yield y
 newtype Reader r e = MkReader (State r e)
   deriving newtype (Handle)
 
+instance (e :> es) => OneWayCoercible (Reader r e) (Reader r es) where
+  oneWayCoercibleImpl = oneWayCoercible
+
 runReader ::
   -- | Initial value for @Reader@.
   r ->
@@ -1398,6 +1526,8 @@ asks ::
   Eff es a
 asks (MkReader st) f = fmap f (get st)
 
+-- | Locally override the value in the @Reader@. It will be restored
+-- when the @local@ block ends.
 local ::
   (e1 :> es) =>
   Reader r e1 ->
@@ -1476,9 +1606,17 @@ runHandleReader h k = do
 
     useImplIn k h'
 
-instance (Handle h) => Handle (HandleReader h) where mapHandle = mapHandleReader
+instance (Handle h) => Handle (HandleReader h) where
+  -- This handleImpl is now a special case that doesn't use
+  -- OneWayCoercion. We can only fix that when we store a
+  -- OneWayCoercion inside `Handle`.
+  handleImpl = handleMapHandle mapHandleReader
 
 newtype ConstEffect r (e :: Effects) = MkConstEffect r
+  deriving (Handle) via OneWayCoercibleHandle (ConstEffect r)
+
+instance (e :> es) => OneWayCoercible (ConstEffect r e) (ConstEffect r es) where
+  oneWayCoercibleImpl = oneWayCoercible
 
 runConstEffect ::
   r ->
@@ -1486,6 +1624,3 @@ runConstEffect ::
   -- | ͘
   Eff es a
 runConstEffect r k = useImplIn k (MkConstEffect r)
-
-instance Handle (ConstEffect r) where
-  mapHandle = coerce
